@@ -4,6 +4,7 @@
 #pragma once
 
 #include "glaze/beve/header.hpp"
+#include "glaze/beve/key_traits.hpp"
 #include "glaze/beve/skip.hpp"
 #include "glaze/core/opts.hpp"
 #include "glaze/core/read.hpp"
@@ -22,7 +23,7 @@ namespace glz
    struct parse<BEVE>
    {
       template <auto Opts, class T, class Tag, is_context Ctx, class It0, class It1>
-         requires(has_no_header(Opts))
+         requires(check_no_header(Opts))
       GLZ_ALWAYS_INLINE static void op(T&& value, Tag&& tag, Ctx&& ctx, It0&& it, It1&& end)
       {
          if constexpr (const_value_v<T>) {
@@ -42,7 +43,7 @@ namespace glz
       }
 
       template <auto Opts, class T, is_context Ctx, class It0, class It1>
-         requires(not has_no_header(Opts))
+         requires(not check_no_header(Opts))
       GLZ_ALWAYS_INLINE static void op(T&& value, Ctx&& ctx, It0&& it, It1&& end)
       {
          if constexpr (const_value_v<T>) {
@@ -73,6 +74,16 @@ namespace glz
          from<BEVE, V>::template op<Opts>(get_member(std::forward<Value>(value), meta_wrapper_v<T>),
                                           std::forward<Ctx>(ctx), std::forward<It0>(it), std::forward<It1>(end));
       }
+
+      template <auto Opts, class Value, class Tag, is_context Ctx, class It0, class It1>
+         requires(check_no_header(Opts))
+      GLZ_ALWAYS_INLINE static void op(Value&& value, Tag&& tag, Ctx&& ctx, It0&& it, It1&& end)
+      {
+         using V = std::decay_t<decltype(get_member(std::declval<Value>(), meta_wrapper_v<T>))>;
+         from<BEVE, V>::template op<Opts>(get_member(std::forward<Value>(value), meta_wrapper_v<T>),
+                                          std::forward<Tag>(tag), std::forward<Ctx>(ctx), std::forward<It0>(it),
+                                          std::forward<It1>(end));
+      }
    };
 
    template <always_null_t T>
@@ -169,7 +180,7 @@ namespace glz
          std::memcpy(data, it, Length);
          it += Length;
 
-         invoke_table<N>([&]<size_t I>() {
+         for_each<N>([&]<size_t I>() {
             get_member(value, get<I>(reflect<T>::values)) = data[I / 8] & (uint8_t{1} << (7 - (I % 8)));
          });
       }
@@ -183,7 +194,7 @@ namespace glz
       static constexpr uint8_t header = tag::number | type | (byte_count<T> << 5);
 
       template <auto Opts>
-         requires(has_no_header(Opts))
+         requires(check_no_header(Opts))
       GLZ_ALWAYS_INLINE static void op(auto&& value, const uint8_t tag, is_context auto&& ctx, auto&& it,
                                        auto&& end) noexcept
       {
@@ -300,7 +311,7 @@ namespace glz
       }
 
       template <auto Opts>
-         requires(not has_no_header(Opts))
+         requires(not check_no_header(Opts))
       GLZ_ALWAYS_INLINE static void op(auto&& value, is_context auto&& ctx, auto&& it, auto&& end) noexcept
       {
          if (invalid_end(ctx, it, end)) {
@@ -321,7 +332,7 @@ namespace glz
       {
          using V = std::underlying_type_t<std::decay_t<T>>;
 
-         if constexpr (has_no_header(Opts)) {
+         if constexpr (check_no_header(Opts)) {
             if ((it + sizeof(V)) > end) [[unlikely]] {
                ctx.error = error_code::unexpected_end;
                return;
@@ -362,7 +373,7 @@ namespace glz
       template <auto Opts>
       static void op(auto&& value, is_context auto&& ctx, auto&& it, auto&& end) noexcept
       {
-         if constexpr (has_no_header(Opts)) {
+         if constexpr (check_no_header(Opts)) {
             using V = std::decay_t<T>;
             if ((it + sizeof(V)) > end) [[unlikely]] {
                ctx.error = error_code::unexpected_end;
@@ -506,7 +517,7 @@ namespace glz
       static_assert(sizeof(V) == 1);
 
       template <auto Opts>
-         requires(has_no_header(Opts))
+         requires(check_no_header(Opts))
       GLZ_ALWAYS_INLINE static void op(auto&& value, const uint8_t, is_context auto&& ctx, auto&& it, auto&& end)
       {
          const auto n = int_from_compressed(ctx, it, end);
@@ -523,7 +534,7 @@ namespace glz
       }
 
       template <auto Opts>
-         requires(not has_no_header(Opts))
+         requires(not check_no_header(Opts))
       GLZ_ALWAYS_INLINE static void op(auto&& value, is_context auto&& ctx, auto&& it, auto&& end)
       {
          constexpr uint8_t header = tag::string;
@@ -1002,9 +1013,7 @@ namespace glz
          using Element = typename T::value_type;
          using Key = typename Element::first_type;
 
-         constexpr uint8_t type = str_t<Key> ? 0 : (std::is_signed_v<Key> ? 0b000'01'000 : 0b000'10'000);
-         constexpr uint8_t byte_cnt = str_t<Key> ? 0 : byte_count<Key>;
-         constexpr uint8_t header = tag::object | type | (byte_cnt << 5);
+         constexpr uint8_t header = beve_key_traits<Key>::header;
 
          if (invalid_end(ctx, it, end)) {
             return;
@@ -1013,7 +1022,7 @@ namespace glz
          if (tag != header) [[unlikely]] {
             if constexpr (check_allow_conversions(Opts)) {
                const auto key_type = tag & 0b000'11'000;
-               if constexpr (str_t<Key>) {
+               if constexpr (beve_key_traits<Key>::as_string) {
                   if (key_type != 0) {
                      ctx.error = error_code::syntax_error;
                      return;
@@ -1040,22 +1049,11 @@ namespace glz
 
          value.clear();
 
-         if constexpr (std::is_arithmetic_v<std::decay_t<Key>>) {
-            constexpr uint8_t key_tag = tag::number | type | (byte_cnt << 5);
-            for (size_t i = 0; i < n; ++i) {
-               // convert the object tag to the key type tag
-               auto& item = value.emplace_back();
-               parse<BEVE>::op<no_header_on<Opts>()>(item.first, key_tag, ctx, it, end);
-               parse<BEVE>::op<Opts>(item.second, ctx, it, end);
-            }
-         }
-         else {
-            constexpr uint8_t key_tag = tag::string;
-            for (size_t i = 0; i < n; ++i) {
-               auto& item = value.emplace_back();
-               parse<BEVE>::op<no_header_on<Opts>()>(item.first, key_tag, ctx, it, end);
-               parse<BEVE>::op<Opts>(item.second, ctx, it, end);
-            }
+         constexpr uint8_t key_tag = beve_key_traits<Key>::key_tag;
+         for (size_t i = 0; i < n; ++i) {
+            auto& item = value.emplace_back();
+            parse<BEVE>::op<no_header_on<Opts>()>(item.first, key_tag, ctx, it, end);
+            parse<BEVE>::op<Opts>(item.second, ctx, it, end);
          }
       }
    };
@@ -1068,9 +1066,7 @@ namespace glz
       {
          using Key = typename T::first_type;
 
-         constexpr uint8_t type = str_t<Key> ? 0 : (std::is_signed_v<Key> ? 0b000'01'000 : 0b000'10'000);
-         constexpr uint8_t byte_cnt = str_t<Key> ? 0 : byte_count<Key>;
-         constexpr uint8_t header = tag::object | type | (byte_cnt << 5);
+         constexpr uint8_t header = beve_key_traits<Key>::header;
 
          if (invalid_end(ctx, it, end)) {
             return;
@@ -1092,7 +1088,7 @@ namespace glz
             return;
          }
 
-         constexpr uint8_t key_tag = type == 0 ? tag::string : (tag::number | type | (byte_cnt << 5));
+         constexpr uint8_t key_tag = beve_key_traits<Key>::key_tag;
          parse<BEVE>::op<no_header_on<Opts>()>(value.first, key_tag, ctx, it, end);
          parse<BEVE>::op<Opts>(value.second, ctx, it, end);
       }
@@ -1106,9 +1102,7 @@ namespace glz
       {
          using Key = typename T::key_type;
 
-         constexpr uint8_t type = str_t<Key> ? 0 : (std::is_signed_v<Key> ? 0b000'01'000 : 0b000'10'000);
-         constexpr uint8_t byte_cnt = str_t<Key> ? 0 : byte_count<Key>;
-         constexpr uint8_t header = tag::object | type | (byte_cnt << 5);
+         constexpr uint8_t header = beve_key_traits<Key>::header;
 
          if (invalid_end(ctx, it, end)) {
             return;
@@ -1117,7 +1111,7 @@ namespace glz
          if (tag != header) [[unlikely]] {
             if constexpr (check_allow_conversions(Opts)) {
                const auto key_type = tag & 0b000'11'000;
-               if constexpr (str_t<Key>) {
+               if constexpr (beve_key_traits<Key>::as_string) {
                   if (key_type != 0) {
                      ctx.error = error_code::syntax_error;
                      return;
@@ -1146,8 +1140,9 @@ namespace glz
             n = value.size();
          }
 
-         if constexpr (std::is_arithmetic_v<std::decay_t<Key>>) {
-            constexpr uint8_t key_tag = tag::number | type | (byte_cnt << 5);
+         constexpr uint8_t key_tag = beve_key_traits<Key>::key_tag;
+
+         if constexpr (beve_key_traits<Key>::as_number) {
             Key key;
             for (size_t i = 0; i < n; ++i) {
                if constexpr (Opts.partial_read) {
@@ -1164,7 +1159,6 @@ namespace glz
             }
          }
          else {
-            constexpr uint8_t key_tag = tag::string;
             static thread_local Key key;
             for (size_t i = 0; i < n; ++i) {
                if constexpr (Opts.partial_read) {
@@ -1236,13 +1230,49 @@ namespace glz
       }
    };
 
+   template <class T>
+      requires(nullable_value_t<T> && not nullable_like<T> && not is_expected<T> && not custom_read<T>)
+   struct from<BEVE, T> final
+   {
+      template <auto Opts>
+      GLZ_ALWAYS_INLINE static void op(auto&& value, is_context auto&& ctx, auto&& it, auto&& end)
+      {
+         if (invalid_end(ctx, it, end)) {
+            return;
+         }
+         const auto tag = uint8_t(*it);
+
+         if (tag == tag::null) {
+            ++it;
+            if constexpr (requires { value.reset(); }) {
+               value.reset();
+            }
+         }
+         else {
+            if (not value.has_value()) {
+               if constexpr (constructible<T>) {
+                  value = meta_construct_v<T>();
+               }
+               else if constexpr (requires { value.emplace(); }) {
+                  value.emplace();
+               }
+               else {
+                  ctx.error = error_code::invalid_nullable_read;
+                  return;
+               }
+            }
+            parse<BEVE>::op<Opts>(value.value(), ctx, it, end);
+         }
+      }
+   };
+
    template <is_includer T>
    struct from<BEVE, T>
    {
       template <auto Opts>
       GLZ_ALWAYS_INLINE static void op(auto&&, is_context auto&& ctx, auto&& it, auto&& end) noexcept
       {
-         if constexpr (has_no_header(Opts)) {
+         if constexpr (check_no_header(Opts)) {
             skip_compressed_int(ctx, it, end);
          }
          else {
@@ -1264,7 +1294,7 @@ namespace glz
    };
 
    template <class T>
-      requires(glaze_object_t<T> || reflectable<T>)
+      requires((glaze_object_t<T> || reflectable<T>) && not custom_read<T>)
    struct from<BEVE, T> final
    {
       template <auto Opts>
@@ -1272,8 +1302,28 @@ namespace glz
       static void op(auto&& value, is_context auto&& ctx, auto&& it, auto&& end)
       {
          if constexpr (reflectable<T>) {
-            auto t = to_tie(value);
-            parse<BEVE>::op<Opts>(t, ctx, it, end);
+            constexpr auto N = detail::count_members<T>;
+            if constexpr (N == 0) {
+               // Handle empty structs by just reading and validating the generic_array header
+               const auto tag = uint8_t(*it);
+               if (tag != tag::generic_array) [[unlikely]] {
+                  ctx.error = error_code::syntax_error;
+                  return;
+               }
+               ++it;
+               const auto n = int_from_compressed(ctx, it, end);
+               if (bool(ctx.error)) [[unlikely]] {
+                  return;
+               }
+               if (n != 0) {
+                  ctx.error = error_code::syntax_error;
+                  return;
+               }
+            }
+            else {
+               auto t = to_tie(value);
+               parse<BEVE>::op<Opts>(t, ctx, it, end);
+            }
          }
          else {
             const auto tag = uint8_t(*it);
@@ -1293,7 +1343,7 @@ namespace glz
                return;
             }
 
-            invoke_table<N>(
+            for_each<N>(
                [&]<size_t I>() { parse<BEVE>::op<Opts>(get_member(value, get<I>(reflect<V>::values)), ctx, it, end); });
          }
       }
@@ -1369,7 +1419,7 @@ namespace glz
                   const sv key{it, n};
                   it += n;
 
-                  jump_table<N>(
+                  visit<N>(
                      [&]<size_t I>() {
                         static constexpr auto TargetKey = get<I>(reflect<T>::keys);
                         static constexpr auto Length = TargetKey.size();
@@ -1452,7 +1502,7 @@ namespace glz
             return;
          }
 
-         invoke_table<N>(
+         for_each<N>(
             [&]<size_t I>() { parse<BEVE>::op<Opts>(get_member(value, get<I>(reflect<T>::values)), ctx, it, end); });
       }
    };
@@ -1483,7 +1533,7 @@ namespace glz
             }
 
             if constexpr (is_std_tuple<T>) {
-               for_each_short_circuit<N>([&](auto I) {
+               for_each_short_circuit<N>([&]<auto I>() {
                   if (I < n) {
                      parse<BEVE>::op<Opts>(std::get<I>(value), ctx, it, end);
                      return false; // continue
@@ -1492,7 +1542,7 @@ namespace glz
                });
             }
             else {
-               for_each_short_circuit<N>([&](auto I) {
+               for_each_short_circuit<N>([&]<auto I>() {
                   if (I < n) {
                      parse<BEVE>::op<Opts>(glz::get<I>(value), ctx, it, end);
                      return false; // continue
@@ -1512,10 +1562,10 @@ namespace glz
             }
 
             if constexpr (is_std_tuple<T>) {
-               invoke_table<N>([&]<size_t I>() { parse<BEVE>::op<Opts>(std::get<I>(value), ctx, it, end); });
+               for_each<N>([&]<size_t I>() { parse<BEVE>::op<Opts>(std::get<I>(value), ctx, it, end); });
             }
             else {
-               invoke_table<N>([&]<size_t I>() { parse<BEVE>::op<Opts>(glz::get<I>(value), ctx, it, end); });
+               for_each<N>([&]<size_t I>() { parse<BEVE>::op<Opts>(glz::get<I>(value), ctx, it, end); });
             }
          }
       }
@@ -1536,15 +1586,13 @@ namespace glz
       }
    };
 
-   template <class T, class Buffer>
-      requires(read_supported<BEVE, T>)
+   template <read_supported<BEVE> T, class Buffer>
    [[nodiscard]] inline error_ctx read_beve(T&& value, Buffer&& buffer)
    {
       return read<opts{.format = BEVE}>(value, std::forward<Buffer>(buffer));
    }
 
-   template <class T, class Buffer>
-      requires(read_supported<BEVE, T>)
+   template <read_supported<BEVE> T, class Buffer>
    [[nodiscard]] inline expected<T, error_ctx> read_beve(Buffer&& buffer)
    {
       T value{};
@@ -1555,8 +1603,7 @@ namespace glz
       return value;
    }
 
-   template <auto Opts = opts{}, class T>
-      requires(read_supported<BEVE, T>)
+   template <auto Opts = opts{}, read_supported<BEVE> T>
    [[nodiscard]] inline error_ctx read_file_beve(T& value, const sv file_name, auto&& buffer)
    {
       context ctx{};
@@ -1571,16 +1618,14 @@ namespace glz
       return read<set_beve<Opts>()>(value, buffer, ctx);
    }
 
-   template <class T, class Buffer>
-      requires(read_supported<BEVE, T>)
+   template <read_supported<BEVE> T, class Buffer>
    [[nodiscard]] inline error_ctx read_binary_untagged(T&& value, Buffer&& buffer)
    {
       return read<opts{.format = BEVE, .structs_as_arrays = true}>(std::forward<T>(value),
                                                                    std::forward<Buffer>(buffer));
    }
 
-   template <class T, class Buffer>
-      requires(read_supported<BEVE, T>)
+   template <read_supported<BEVE> T, class Buffer>
    [[nodiscard]] inline expected<T, error_ctx> read_binary_untagged(Buffer&& buffer)
    {
       T value{};
@@ -1591,8 +1636,7 @@ namespace glz
       return value;
    }
 
-   template <auto Opts = opts{}, class T>
-      requires(read_supported<BEVE, T>)
+   template <auto Opts = opts{}, read_supported<BEVE> T>
    [[nodiscard]] inline error_ctx read_file_beve_untagged(T& value, const std::string& file_name, auto&& buffer)
    {
       return read_file_beve<opt_true<Opts, &opts::structs_as_arrays>>(value, file_name, buffer);
